"""Helpers and utils."""
from __future__ import annotations

from const import REGIONS, UNKNOWN


def get_region_name(region_id: int) -> str:
    region_name = REGIONS.get(region_id, f'{UNKNOWN} <{region_id}>')
    return region_name



"""Taipit API wrapper."""
from __future__ import annotations

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")

from .api import TaipitApi
from .auth import AbstractTaipitAuth, SimpleTaipitAuth

__all__ = [
    "TaipitApi",
    "AbstractTaipitAuth",
    "SimpleTaipitAuth",
    __version__
]



"""Provide a CLI for Taipit."""
from __future__ import annotations

import argparse
import logging
from pprint import pprint

from requests import Session

from ._version import __version__
from .api import TaipitApi
from .auth import SimpleTaipitAuth
from .const import (
    DEFAULT_CLIENT_ID,
    DEFAULT_CLIENT_SECRET,
    GUEST_USERNAME,
    GUEST_PASSWORD, LOG_LEVELS
)


def get_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Command line tool for Taipit API")
    parser.add_argument("-u", "--username", help="User name")
    parser.add_argument("-p", "--password", help="Password")
    parser.add_argument("--client_id", required=False,
                        help="OAuth credentials client_id (Optional.)",
                        default=DEFAULT_CLIENT_ID)
    parser.add_argument("--client_secret", required=False,
                        help="OAuth credentials client_secret (Optional.)",
                        default=DEFAULT_CLIENT_SECRET)

    # command

    parser.add_argument('id',
                        nargs="?",
                        help='meter ID, if not specified, '
                             'information about all meters will be shown')

    parser.add_argument('--readings',
                        help='show readings for meter', action="store_true")
    parser.add_argument('--info',
                        help='show information about meter', action="store_true")
    parser.add_argument("--user",
                        nargs="?",
                        dest='user_id',
                        help="show info about user, if user_id not specified, "
                             "information about current user will be shown")

    # warnings
    parser.add_argument("--warnings",
                        help="show warnings from the Taipit",
                        action="store_true")
    # settings
    parser.add_argument("--settings",
                        help="show settings for Taipit API",
                        action="store_true")

    parser.add_argument('-v', '--verbose',
                        action='count', default=0,
                        help="increase verbosity level")
    parser.add_argument("-V", "--version",
                        action="version", version=__version__)

    arguments = parser.parse_args()

    return arguments


def cli() -> None:
    """Run main."""
    args = get_arguments()

    # Setup logging and the log level according to the "-v" option
    logging.basicConfig(level=LOG_LEVELS.get(args.verbose, logging.INFO))

    if not args.username:
        username = input(f"User name (default: {GUEST_USERNAME}):")
        if not username:
            username = GUEST_USERNAME
            password = input(f"Password (default: {GUEST_PASSWORD}):")
            if not password:
                password = GUEST_PASSWORD
        else:
            password = input("Password:")
    else:
        username = args.username
        password = args.password

    with Session() as session:
        auth = SimpleTaipitAuth(username, password, session, args.client_id, args.client_secret)
        api = TaipitApi(auth)

        if args.info:
            if args.id:
                print(f"Info about Meter ID={args.id}:")
                _meter_info = api.get_meter_info(args.id)
                pprint(_meter_info)
            else:
                print("Info about all meters:")
                _meters = api.get_own_meters()
                pprint(_meters)
            return

        if args.user_id:
            print(f"Info about User ID={args.user_id}:")
            _user = api.get_user_info(args.user_id)
            pprint(_user)
        else:
            print("Current User Info:")
            _current_user = api.get_current_user()
            pprint(_current_user)

        if args.settings:
            print("Settings:")
            _settings = api.get_settings()
            pprint(_settings)

        if args.warnings:
            print("Warnings:")
            _warnings = api.get_warnings()
            pprint(_warnings)

        if args.readings:
            if args.id:
                print(f"Readings Meter ID={args.id}:")
                _readings = api.get_meter_readings(args.id)
                pprint(_readings)

        # show all reading in all other cases
        print("Readings:")
        _all_readings = api.get_meters()
        pprint(_all_readings)


if __name__ == "__main__":
    cli()



import pytest
from requests import Session

from pytaipit import SimpleTaipitAuth, TaipitApi
from .common import (
    TEST_USERNAME,
    TEST_PASSWORD,
    TEST_CLIENT_ID,
    TEST_CLIENT_SECRET
)


@pytest.fixture(scope='class')
def auth() -> SimpleTaipitAuth:
    with Session() as _session:
        _auth = SimpleTaipitAuth(
            username=TEST_USERNAME,
            password=TEST_PASSWORD,
            session=_session,
            client_id=TEST_CLIENT_ID,
            client_secret=TEST_CLIENT_SECRET
        )
        yield _auth
        # some finalization


@pytest.fixture(scope='class')
def api(auth: SimpleTaipitAuth) -> TaipitApi:
    _api = TaipitApi(auth)
    yield _api
    # some finalization



TEST_CLIENT_ID = '1_34mi8uiv22iowgsc0wggk0c8888cc80s8gk80gco00g0gko8ko'
TEST_CLIENT_SECRET = '15lkcj9ytmm8440ggsco8ogo4ockwgcg04okokcwokkk8cksk8'
TEST_USERNAME = "guest@taipit.ru"  # may be changed to real username
TEST_PASSWORD = "guest"  # change accordingly TEST_USERNAME



import pytest

from pytaipit import TaipitApi
from pytaipit.const import (
    SECTION_CONTROLLERS,
    SECTION_METER_TYPES,
    SECTION_REGIONS, SECTION_METER_TYPES_FULL
)

CONF_CONTROLLER_ID = 'controllerId'
CONF_ENERGY_T3_A = 'energy_t3_a'
CONF_ENERGY_T2_A = 'energy_t2_a'
CONF_ENERGY_T1_A = 'energy_t1_a'
CONF_ENERGY_A = 'energy_a'
CONF_LAST_READING = 'lastReading'
CONF_ECOMETER_DATA = 'ecometerdata'
CONF_METER_NAME = 'metername'
CONF_SUCCESS = 'success'
CONF_DATA = 'data'
CONF_USERNAME = 'username'
CONF_ID = 'id'


@pytest.mark.asyncio
class TestTaipitApi:
    @pytest.fixture(scope='class')
    def user(self) -> dict:
        return {}

    @pytest.fixture(scope='class')
    def meter_ids(self) -> list:
        return []

    def test_get_meters(self, api: TaipitApi, meter_ids: list):
        _meters = api.get_meters()
        assert _meters is not None
        assert isinstance(_meters, list)
        assert len(_meters) > 0
        for meter in _meters:
            assert CONF_ID in meter
            assert CONF_METER_NAME in meter
            assert CONF_ECOMETER_DATA in meter
            assert CONF_LAST_READING in meter[CONF_ECOMETER_DATA]
            assert CONF_ENERGY_A in meter[CONF_ECOMETER_DATA][CONF_LAST_READING]
            assert CONF_ENERGY_T1_A in meter[CONF_ECOMETER_DATA][CONF_LAST_READING]
            assert CONF_ENERGY_T2_A in meter[CONF_ECOMETER_DATA][CONF_LAST_READING]
            assert CONF_ENERGY_T3_A in meter[CONF_ECOMETER_DATA][CONF_LAST_READING]
            meter_ids.append(meter[CONF_ID])

    def test_get_meter_readings(self, api: TaipitApi, meter_ids: list):
        for meter_id in meter_ids:
            _readings = api.get_meter_readings(meter_id)

    def test_get_own_meters(self, api: TaipitApi, meter_ids: list):
        _meters = api.get_own_meters()
        assert _meters is not None
        assert isinstance(_meters, list)
        for meter in _meters:
            if meter[CONF_ID] not in meter_ids:
                meter_ids.append(meter[CONF_ID])

    def test_get_meter_info(self, api: TaipitApi, meter_ids: list):
        for meter_id in meter_ids:
            _info = api.get_meter_info(meter_id)
            assert _info
            assert CONF_ID in _info
            assert meter_id == _info[CONF_ID]

    def test_get_current_user(self, api: TaipitApi, user: dict):
        _current_user = api.get_current_user()
        assert _current_user
        assert CONF_USERNAME in _current_user
        assert CONF_ID in _current_user
        user.update(_current_user)

    def test_get_user_info(self, api: TaipitApi, user):
        _user_info = api.get_user_info(user[CONF_ID])

    def test_get_warnings(self, api: TaipitApi):
        _warnings = api.get_warnings()
        assert _warnings is not None
        assert CONF_DATA in _warnings
        assert isinstance(_warnings[CONF_DATA], list)
        assert CONF_SUCCESS in _warnings
        assert CONF_SUCCESS in _warnings
        assert _warnings[CONF_SUCCESS]

    def test_get_settings(self, api: TaipitApi):
        _settings = api.get_settings()
        assert _settings is not None
        assert SECTION_CONTROLLERS in _settings
        assert isinstance(_settings[SECTION_CONTROLLERS], list)
        assert SECTION_METER_TYPES in _settings
        assert isinstance(_settings[SECTION_METER_TYPES], list)
        assert SECTION_METER_TYPES_FULL in _settings
        assert isinstance(_settings[SECTION_METER_TYPES_FULL], list)
        assert SECTION_REGIONS in _settings
        assert isinstance(_settings[SECTION_REGIONS], list)



import pytest


@pytest.mark.asyncio
class TestTaipitAuth:
    def test_new_token(self, auth):
        token = auth._new_token()
        assert 'access_token' in token
        assert 'expires_in' in token
        assert 'expires_at' in token
        assert 'refresh_token' in token
        assert 'token_type' in token
        assert 'scope' in token
        assert token['scope'] is None
        assert token['token_type'] == 'bearer'
        assert token['expires_in'] == 3600
        auth._token = token

    def test_valid_token(self, auth):
        assert auth._is_valid_token(auth._token) is True

    def test_refresh_token(self, auth):
        token = auth._refresh_token(auth._token)
        assert 'access_token' in token
        assert 'expires_in' in token
        assert 'expires_at' in token
        assert 'refresh_token' in token
        assert 'token_type' in token
        assert 'scope' in token
        assert token['scope'] is None
        assert token['token_type'] == 'bearer'
        assert token['expires_in'] == 3600
        auth._token = token



"""Taipit API wrapper."""
from __future__ import annotations

from typing import List, Any, Union

from .auth import AbstractTaipitAuth
from .const import DEFAULT_API_URL, SECTIONS_ALL, \
    PARAM_ACTION, PARAM_SECTIONS, GET_ENTRIES, PARAM_ID


class TaipitApi:
    """Class to communicate with the Taipit API."""
    _api_url: str

    def __init__(self, auth: AbstractTaipitAuth, api_url: str = DEFAULT_API_URL):
        """Initialize the API and store the auth."""
        self._auth = auth
        self._api_url = api_url

    def get(self, url: str, **kwargs) -> Union[dict[str, Any], list[dict[str, Any]]]:
        """Make async get request to api endpoint"""
        return self._auth.request("GET", f"{self._api_url}/{url}", **kwargs)

    def get_meters(self) -> list[dict[str, Any]]:
        """Get all meters and short info."""
        _url = 'meter/list-all'
        return self.get(_url)

    def get_meter_readings(self, meter_id: int) -> dict[str, Any]:
        """Get readings for meter."""
        _url = 'bmd/all'
        params = {PARAM_ID: meter_id}
        return self.get(_url, params=params)

    def get_own_meters(self) -> list[dict[str, Any]]:
        """Get meters owned by current user."""
        _url = 'meter/list-owner'
        return self.get(_url)

    def get_meter_info(self, meter_id: int) -> dict[str, Any]:
        """Get info for meter."""
        _url = 'meter/get-id'
        params = {PARAM_ID: meter_id}
        return self.get(_url, params=params)

    def get_current_user(self) -> dict[str, Any]:
        """Get current user info."""
        _url = 'user/getuser'
        return self.get(_url)

    def get_user_info(self, user_id: str) -> dict[str, Any]:
        """Get specified user info."""
        _url = f'user/getuserinfo/{user_id}'
        return self.get(_url)

    def get_warnings(self) -> dict[str, Any]:
        """List warnings."""
        _params = {PARAM_ACTION: GET_ENTRIES}
        _url = f'warnings/list'
        return self.get(_url, params=_params)

    def get_settings(self, sections: List[str] = SECTIONS_ALL) -> dict[str, Any]:
        """Get settings"""
        _params = {PARAM_SECTIONS: ','.join(sections)}
        _url = f'config/settings'
        return self.get(_url, params=_params)



"""Exceptions for Taipit."""
from __future__ import annotations


class TaipitError(Exception):
    """Base class for aiotaipit errors."""


class TaipitAuthError(TaipitError):
    """Base class for aiotaipit errors."""


class TaipitAuthInvalidGrant(TaipitAuthError):
    """Invalid username and password combination"""


class TaipitAuthInvalidClient(TaipitAuthError):
    """The client credentials are invalid"""


class TaipitTokenError(TaipitAuthError):
    """Taipit token error"""


class TaipitInvalidTokenResponse(TaipitTokenError):
    """Invalid token response"""


class TaipitTokenAcquireFailed(TaipitTokenError):
    """Taipit token error"""


class TaipitTokenRefreshFailed(TaipitTokenError):
    """Token refresh failed"""



""" Example showing usage of this library """

from pprint import pprint

import requests

from pytaipit import SimpleTaipitAuth, TaipitApi


def main(username: str, password: str) -> None:
    """Create the aiohttp session and run the example."""
    with requests.Session() as session:
        auth = SimpleTaipitAuth(username, password, session)
        api = TaipitApi(auth)

        meters = api.get_meters()

        pprint(meters)


if __name__ == "__main__":
    _username = "guest@taipit.ru"
    _password = "guest"
    main(_username, _password)



"""Constants for aiotaipit."""
from __future__ import annotations

import logging

LOGGER = logging.getLogger(__package__)

LOG_LEVELS = {
    None: logging.WARNING,  # 0
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}

# OAuth client ID and secret
#
# В конфигурации используются идентификатор клиента и секрет,
# используемые мобильным приложением на момент разработки.
# Если это перестанет работать, вы можете найти новые с помощью прокси-сервера MITM с мобильным приложением.
# Приложение вызовет https://cloud.meters.taipit.ru/oauth/v2/token с client_id и client_secret.
# Затем вы можете использовать эти значения во время настройки.

DEFAULT_CLIENT_ID = '1_34mi8uiv22iowgsc0wggk0c8888cc80s8gk80gco00g0gko8ko'
DEFAULT_CLIENT_SECRET = '15lkcj9ytmm8440ggsco8ogo4ockwgcg04okokcwokkk8cksk8'
GUEST_USERNAME = "guest@taipit.ru"
GUEST_PASSWORD = "guest"

DEFAULT_BASE_URL = "https://cloud.meters.taipit.ru"
DEFAULT_TOKEN_URL = "oauth/v2/token"
DEFAULT_API_URL = "api"

SECTION_CONTROLLERS = 'controllers'
SECTION_METER_TYPES = 'meterTypes'
SECTION_REGIONS = 'regions'
SECTIONS_ALL = (SECTION_REGIONS, SECTION_METER_TYPES, SECTION_CONTROLLERS)

PARAM_ID = 'id'
PARAM_ACTION = 'action'
PARAM_SECTIONS = 'sections'

GET_ENTRIES = 'getEntries'

UNKNOWN = 'Неизвестно'

METER_MODELS = {
    15: 'МТ 315 (GSM)',
    16: 'МТ 114 (Wi-Fi)',
    21: 'МТ 124 (Wi-Fi)',
    22: 'МТ 115 (GSM)',
}

REGIONS = {
    1: 'Республика Адыгея',
    2: 'Республика Башкортостан',
    3: 'Республика Бурятия',
    4: 'Республика Алтай',
    5: 'Республика Дагестан',
    6: 'Республика Ингушетия',
    7: 'Кабардино-Балкарская республика',
    8: 'Республика Калмыкия',
    9: 'Карачаево-Черкесская республика',
    10: 'Республика Карелия',
    11: 'Республика Коми',
    12: 'Республика Марий Эл',
    13: 'Республика Мордовия',
    14: 'Республика Саха (Якутия)',
    15: 'Республика Северная Осетия — Алания',
    16: 'Республика Татарстан',
    17: 'Республика Тыва',
    18: 'Удмуртская республика',
    19: 'Республика Хакасия',
    20: 'Чеченская республика',
    21: 'Чувашская республика',
    22: 'Алтайский край',
    23: 'Краснодарский край',
    24: 'Красноярский край',
    25: 'Приморский край',
    26: 'Ставропольский край',
    27: 'Хабаровский край',
    28: 'Амурская область',
    29: 'Архангельская область',
    30: 'Астраханская область',
    31: 'Белгородская область',
    32: 'Брянская область',
    33: 'Владимирская область',
    34: 'Волгоградская область',
    35: 'Вологодская область',
    36: 'Воронежская область',
    37: 'Ивановская область',
    38: 'Иркутская область',
    39: 'Калининградская область',
    40: 'Калужская область',
    41: 'Камчатский край',
    42: 'Кемеровская область',
    43: 'Кировская область',
    44: 'Костромская область',
    45: 'Курганская область',
    46: 'Курская область',
    47: 'Ленинградская область',
    48: 'Липецкая область',
    49: 'Магаданская область',
    50: 'Московская область',
    51: 'Мурманская область',
    52: 'Нижегородская область',
    53: 'Новгородская область',
    54: 'Новосибирская область',
    55: 'Омская область',
    56: 'Оренбургская область',
    57: 'Орловская область',
    58: 'Пензенская область',
    59: 'Пермский край',
    60: 'Псковская область',
    61: 'Ростовская область',
    62: 'Рязанская область',
    63: 'Самарская область',
    64: 'Саратовская область',
    65: 'Сахалинская область',
    66: 'Свердловская область',
    67: 'Смоленская область',
    68: 'Тамбовская область',
    69: 'Тверская область',
    70: 'Томская область',
    71: 'Тульская область',
    72: 'Тюменская область',
    73: 'Ульяновская область',
    74: 'Челябинская область',
    75: 'Забайкальский край',
    76: 'Ярославская область',
    77: 'Москва',
    78: 'Санкт-Петербург',
    79: 'Еврейская автономная область',
    83: 'Ненецкий автономный округ',
    86: 'Ханты-Мансийский автономный округ - Югра',
    87: 'Чукотский автономный округ',
    89: 'Ямало-Ненецкий автономный округ',
    91: 'Республика Крым',
    92: 'Севастополь'
}
SECTION_METER_TYPES_FULL = 'meterTypesFull'

TOKEN_REQUIRED_FIELDS = {'access_token', 'expires_in', 'refresh_token'}



"""Taipit API Auth wrapper."""
from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import Dict, Any, cast, Optional

from requests import Session

from .const import (
    DEFAULT_CLIENT_ID,
    DEFAULT_CLIENT_SECRET,
    DEFAULT_TOKEN_URL,
    DEFAULT_BASE_URL, TOKEN_REQUIRED_FIELDS, LOGGER
)
from .exceptions import (
    TaipitInvalidTokenResponse,
    TaipitAuthInvalidGrant,
    TaipitAuthInvalidClient,
    TaipitAuthError,
    TaipitTokenRefreshFailed,
    TaipitTokenAcquireFailed
)

CLOCK_OUT_OF_SYNC_MAX_SEC = 20


class AbstractTaipitAuth(ABC):
    """Abstract class to make authenticated requests."""
    _session: Session
    _base_url: str

    def __init__(self, session: Session | None = None, base_url: str = DEFAULT_BASE_URL):
        """Initialize the auth."""
        self._session = session if session else Session()
        self._base_url = base_url

    @abstractmethod
    def get_access_token(self) -> str:
        """Return a valid access token."""

    def request(self, method: str, url: str, **kwargs) -> Any:
        """Make a request with token authorization."""
        _url = f"{self._base_url}/{url}"
        if "headers" not in kwargs:
            kwargs["headers"] = {}
        access_token = self.get_access_token()
        kwargs["headers"]["Authorization"] = f"Bearer {access_token}"

        LOGGER.debug("Request to %s with data %s", url, kwargs)

        response = self._session.request(method, _url, **kwargs)
        response.raise_for_status()
        data = response.json()
        LOGGER.debug("Request finished with status=%s, headers=%s, data=%s",
                     response.status_code, response.headers, data)
        return data


class SimpleTaipitAuth(AbstractTaipitAuth):
    """Simple implementation of AbstractFlickAuth that gets a token once."""
    _username: str
    _password: str
    _client_id: str
    _client_secret: str
    _token_url: str
    _token: Optional[Dict[str, Any]] = None

    def __init__(self, username: str, password: str,
                 session: Session | None = None,
                 client_id: str = DEFAULT_CLIENT_ID,
                 client_secret: str = DEFAULT_CLIENT_SECRET,
                 base_url: str = DEFAULT_BASE_URL,
                 token_url: str = DEFAULT_TOKEN_URL):
        super().__init__(session, base_url)
        self._username = username
        self._password = password
        self._client_id = client_id
        self._client_secret = client_secret
        self._token_url = token_url

    def _token_request(self, data: dict) -> dict:
        """Make a token request."""
        _url = f"{self._base_url}/{self._token_url}"
        data["client_id"] = self._client_id
        data["client_secret"] = self._client_secret

        LOGGER.debug("Token request %s with data %s",
                     _url, data)

        response = self._session.get(_url, params=data)

        if response.status_code == 400:
            error_info = response.json()
            if error_info['error'] == 'invalid_grant':
                raise TaipitAuthInvalidGrant(error_info['error_description'])
            if error_info['error'] == 'invalid_client':
                raise TaipitAuthInvalidClient(error_info['error_description'])
            raise TaipitAuthError

        response.raise_for_status()

        new_token = response.json()

        if not self._is_valid_token(new_token):
            raise TaipitInvalidTokenResponse

        new_token["expires_in"] = int(new_token["expires_in"])
        new_token["expires_at"] = time.time() + new_token["expires_in"]

        LOGGER.debug("Token request finished %s", new_token)

        return cast(dict, new_token)

    def _refresh_token(self, token: dict) -> dict:
        """Refresh tokens."""
        try:
            new_token = self._token_request(
                {
                    "grant_type": "refresh_token",
                    "refresh_token": token["refresh_token"],
                }
            )
        except (TaipitAuthInvalidGrant, TaipitAuthInvalidClient):
            raise
        except Exception as exc:
            raise TaipitTokenRefreshFailed from exc

        return cast(dict, new_token)

    def _new_token(self) -> dict:
        """Refresh tokens."""
        try:
            new_token = self._token_request(
                {
                    "grant_type": "password",
                    "username": self._username,
                    "password": self._password
                }
            )

        except (TaipitAuthInvalidGrant, TaipitAuthInvalidClient):
            raise
        except Exception as exc:
            raise TaipitTokenAcquireFailed from exc

        return cast(dict, new_token)

    @staticmethod
    def _is_valid_token(token: dict) -> bool:
        """Check if token is valid and contains all required fields."""
        return TOKEN_REQUIRED_FIELDS <= token.keys()

    @staticmethod
    def _is_expired_token(token: dict) -> bool:
        """Check if token is not expired"""
        return (
                cast(float, token["expires_at"])
                < time.time() + CLOCK_OUT_OF_SYNC_MAX_SEC
        )

    def get_access_token(self):
        """Get access token"""
        if self._token:
            if self._is_expired_token(self._token):
                self._token = self._refresh_token(self._token)
        else:
            self._token = self._new_token()

        return self._token["access_token"]
